﻿Public Class Form1

    Dim r As IO.BinaryReader

    Dim Patterns As New List(Of Integer)
    Dim Meshes As New List(Of Mesh)

    Enum EVersion
        U0864v
        U0867z
        U0874d
    End Enum

    Dim Version As EVersion

    Sub Log(text As String)
        lbLog.Items.Add(text)
    End Sub

    Structure Mesh
        Dim Address As Integer
        Dim Verts, Frames, Polys As Integer
        Dim PolyData, AnimData As Byte()
    End Structure

    Sub SearchForPatterns()
        While r.BaseStream.Length - r.BaseStream.Position > 141
            If CheckPattern() Then
                Log("Found pattern at " & r.BaseStream.Position)
                Patterns.Add(r.BaseStream.Position)
            End If

            If (r.BaseStream.Position And 32767) = 0 Then
                pbProgress.Value = r.BaseStream.Position / (r.BaseStream.Length - 141) * 100
            End If
        End While
    End Sub

    Sub TryToReadMesh(patNum As Integer)
        Dim pos = Patterns(patNum)
        Log("Trying at " & pos)
        r.BaseStream.Position = pos

        Try
            Dim vertCount = r.ReadUInt32
            Dim frameCount = r.ReadUInt32

            If r.ReadUInt32 <> 65535 Then
                Log("No FF FF")
                Return
            End If

            Dim offset = If(Version <= EVersion.U0864v, 42, 48)

            Dim polyDefStart = r.BaseStream.Position + offset

            Do
                If Check8F() Then
                    Exit Do
                End If
            Loop

            Dim polyDefLen = r.BaseStream.Position - 8 - polyDefStart

            'If polyDefLen Mod 16 <> 0 Then
            '    Log("Size is not multiply of 16 (" & polyDefLen Mod 16 & ")")
            '    Return
            'End If

            r.BaseStream.Position = polyDefStart

            Dim polyDefData = r.ReadBytes(polyDefLen)

            r.BaseStream.Position += 8

            Dim polyCount = r.ReadUInt16
            Log("Data size: " & polyDefLen & ", expected: " & 16 * polyCount)
            'If polyCount * 16 <> polyDefLen Then
            '    Log("Size and polycount don't match")
            '    Return
            'End If

            'search for anim data start
            'If Version < EVersion.U0874d Then
            '    If Version = EVersion.U0867z Then
            '        Do
            '            If Check8F() Then
            '                Exit Do
            '            End If
            '        Loop
            '    End If

            '    Do
            '        If Check0101() Then
            '            Exit Do
            '        End If
            '    Loop
            'Else
            '    Do
            '        If Check8F() Then
            '            Exit Do
            '        End If
            '    Loop
            '    r.BaseStream.Position += 12
            'End If
            If Version < EVersion.U0874d Then
                Do
                    If Check8F() Then
                        Exit Do
                    End If
                Loop
                Do
                    If Check0101() Then
                        Exit Do
                    End If
                Loop
            Else
                Do
                    If Check8F() Then
                        Exit Do
                    End If
                Loop
                r.BaseStream.Position += 12
            End If

            Dim animData = r.ReadBytes(4 * vertCount * frameCount)

            Dim newMesh As New Mesh

            newMesh.Address = pos
            newMesh.Verts = vertCount
            newMesh.Frames = frameCount
            newMesh.Polys = polyCount
            newMesh.PolyData = polyDefData
            newMesh.AnimData = animData

            Meshes.Add(newMesh)
            Log("Success")

        Catch e As IO.EndOfStreamException
            Log("End of stream error")
        End Try

    End Sub

    Sub ScanFile(file As String)
        lbLog.Items.Clear()
        Patterns.Clear()
        Meshes.Clear()
        'this way it will be faster
        r = New IO.BinaryReader(New IO.MemoryStream(IO.File.ReadAllBytes(file)))

        SearchForPatterns()

        For i = 0 To Patterns.Count - 1
            TryToReadMesh(i)
        Next

        r.Close()

    End Sub

    Function CheckPattern() As Boolean
        Dim returnTo = r.BaseStream.Position + 1

        For i = 0 To 7
            If r.ReadByte <> 255 Then
                GoTo bad
            End If
        Next

        r.BaseStream.Position += If(Version >= EVersion.U0874d, 53, 41)

        Dim x = r.ReadUInt32
        If r.ReadUInt32 <> x - 2 Then GoTo bad
        If r.ReadUInt32 <> x - 1 Then GoTo bad
        If r.ReadUInt32 <> x + 1 Then GoTo bad
        If r.ReadUInt32 <> x + 2 Then GoTo bad
        If r.ReadUInt32 <> x + 3 Then GoTo bad
        If r.ReadUInt32 <> x + 4 Then GoTo bad
        If r.ReadUInt32 <> x + 5 Then GoTo bad

        'stops at vertcount
        Return True
bad:
        r.BaseStream.Position = returnTo
        Return False
    End Function

    Function Check8F() As Boolean
        Dim returnTo = r.BaseStream.Position + 1

        For i = 0 To 7
            If r.ReadByte <> 255 Then
                r.BaseStream.Position = returnTo
                Return False
            End If
        Next

        Return True
    End Function

    Function Check0101() As Boolean
        Dim returnTo = r.BaseStream.Position + 1

        If r.ReadByte <> 1 Then GoTo bad
        If r.ReadByte <> 1 Then GoTo bad
        If r.ReadByte <> 0 Then GoTo bad
        If r.ReadByte <> 0 Then GoTo bad
        Return True
bad:
        r.BaseStream.Position = returnTo
        Return False
    End Function

    Private Sub btnOpen_Click(sender As System.Object, e As System.EventArgs) Handles btnOpen.Click
        If OpenDlg.ShowDialog = Windows.Forms.DialogResult.OK Then
            ScanFile(OpenDlg.FileName)
        End If
    End Sub

    Private Sub btnExport_Click(sender As System.Object, e As System.EventArgs) Handles btnExport.Click
        If FolderDlg.ShowDialog = Windows.Forms.DialogResult.OK Then
            For i = 0 To Meshes.Count - 1
                Dim file = IO.Path.Combine(FolderDlg.SelectedPath, "Mesh" & i)
                Dim fileA = file & "_a.3d"
                Dim fileD = file & "_d.3d"

                Dim m = Meshes(i)

                Dim wD As New IO.BinaryWriter(New IO.FileStream(fileD, IO.FileMode.CreateNew))
                wD.Write(CUShort(m.Polys))
                wD.Write(CUShort(m.Verts))
                wD.Seek(44, IO.SeekOrigin.Current)
                wD.Write(m.PolyData)
                wD.Close()

                Dim wA As New IO.BinaryWriter(New IO.FileStream(fileA, IO.FileMode.CreateNew))
                wA.Write(CUShort(m.Frames))
                wA.Write(CUShort(m.Verts * 4))
                wA.Write(m.AnimData)
                wA.Close()

            Next
        End If
    End Sub

    Private Sub rb864v_CheckedChanged(sender As System.Object, e As System.EventArgs) Handles rb864v.CheckedChanged
        If rb864v.Checked Then
            Version = EVersion.U0864v
        End If
    End Sub

    Private Sub rb867z_CheckedChanged(sender As System.Object, e As System.EventArgs) Handles rb867z.CheckedChanged
        If rb867z.Checked Then
            Version = EVersion.U0867z
        End If
    End Sub

    Private Sub rb874d_CheckedChanged(sender As System.Object, e As System.EventArgs) Handles rb874d.CheckedChanged
        If rb874d.Checked Then
            Version = EVersion.U0874d
        End If
    End Sub
End Class